import pytest

from ecoinvent_migrate.errors import Mismatch, Uncombinable
from ecoinvent_migrate.wrangling import source_target_pair_as_bw_dict


def test_source_target_pair_as_bw_dict_simple():
    given = {
        "Activity Name - 3.9.1": "baling",
        "Geography - 3.9.1": "GLO",
        "Reference Product - 3.9.1": "baling",
        "Reference Product Unit - 3.9.1": "unit",
        "Activity Name - 3.10": "baling",
        "Geography - 3.10": "GLO",
        "Reference Product - 3.10": "baling",
        "Reference Product Unit - 3.10": "unit",
    }
    expected = [
        {
            "source": {
                "name": "baling",
                "location": "GLO",
                "reference product": "baling",
                "unit": "unit",
            },
            "target": {
                "name": "baling",
                "location": "GLO",
                "reference product": "baling",
                "unit": "unit",
            },
        }
    ]
    assert source_target_pair_as_bw_dict(given, "3.9.1", "3.10") == expected


def test_source_target_pair_as_bw_dict_multiple():
    given = {
        "Activity Name - 3.9.1": "autoclaved aerated concrete block production",
        "Geography - 3.9.1": "IN",
        "Reference Product - 3.9.1": "autoclaved aerated concrete block;\nhard coal ash",
        "Reference Product Unit - 3.9.1": "kg;\nkg",
        "Activity Name - 3.10": "autoclaved aerated concrete block production",
        "Geography - 3.10": "IN",
        "Reference Product - 3.10": "autoclaved aerated concrete block;\nhard coal ash",
        "Reference Product Unit - 3.10": "kg;\nkg",
    }
    expected = [
        {
            "source": {
                "name": "autoclaved aerated concrete block production",
                "location": "IN",
                "reference product": "autoclaved aerated concrete block",
                "unit": "kg",
            },
            "target": {
                "name": "autoclaved aerated concrete block production",
                "location": "IN",
                "reference product": "autoclaved aerated concrete block",
                "unit": "kg",
            },
        },
        {
            "source": {
                "name": "autoclaved aerated concrete block production",
                "location": "IN",
                "reference product": "hard coal ash",
                "unit": "kg",
            },
            "target": {
                "name": "autoclaved aerated concrete block production",
                "location": "IN",
                "reference product": "hard coal ash",
                "unit": "kg",
            },
        },
    ]
    assert source_target_pair_as_bw_dict(given, "3.9.1", "3.10") == expected


def test_source_target_pair_as_bw_dict_n_to_1():
    given = {
        "Activity Name - 3.9.1": "autoclaved aerated concrete block production",
        "Geography - 3.9.1": "IN",
        "Reference Product - 3.9.1": "autoclaved aerated concrete block;\nhard coal ash",
        "Reference Product Unit - 3.9.1": "kg;\nkg",
        "Activity Name - 3.10": "autoclaved aerated concrete block production",
        "Geography - 3.10": "IN",
        "Reference Product - 3.10": "autoclaved aerated concrete block",
        "Reference Product Unit - 3.10": "kg",
    }
    expected = [
        {
            "source": {
                "name": "autoclaved aerated concrete block production",
                "location": "IN",
                "reference product": "autoclaved aerated concrete block",
                "unit": "kg",
            },
            "target": {
                "name": "autoclaved aerated concrete block production",
                "location": "IN",
                "reference product": "autoclaved aerated concrete block",
                "unit": "kg",
            },
        },
        {
            "source": {
                "name": "autoclaved aerated concrete block production",
                "location": "IN",
                "reference product": "hard coal ash",
                "unit": "kg",
            },
            "target": {
                "name": "autoclaved aerated concrete block production",
                "location": "IN",
                "reference product": "autoclaved aerated concrete block",
                "unit": "kg",
            },
        },
    ]
    assert source_target_pair_as_bw_dict(given, "3.9.1", "3.10") == expected


def test_source_target_pair_as_bw_dict_1_to_n():
    given = {
        "Activity Name - 3.9.1": "autoclaved aerated concrete block production",
        "Geography - 3.9.1": "IN",
        "Reference Product - 3.9.1": "autoclaved aerated concrete block",
        "Reference Product Unit - 3.9.1": "kg",
        "Activity Name - 3.10": "autoclaved aerated concrete block production",
        "Geography - 3.10": "IN",
        "Reference Product - 3.10": "autoclaved aerated concrete block;\nhard coal ash",
        "Reference Product Unit - 3.10": "kg;\nkg",
    }
    expected = [
        {
            "source": {
                "name": "autoclaved aerated concrete block production",
                "location": "IN",
                "reference product": "autoclaved aerated concrete block",
                "unit": "kg",
            },
            "target": {
                "name": "autoclaved aerated concrete block production",
                "location": "IN",
                "reference product": "autoclaved aerated concrete block",
                "unit": "kg",
            },
        },
        {
            "source": {
                "name": "autoclaved aerated concrete block production",
                "location": "IN",
                "reference product": "autoclaved aerated concrete block",
                "unit": "kg",
            },
            "target": {
                "name": "autoclaved aerated concrete block production",
                "location": "IN",
                "reference product": "hard coal ash",
                "unit": "kg",
            },
        },
    ]
    assert source_target_pair_as_bw_dict(given, "3.9.1", "3.10") == expected


def test_source_target_pair_as_bw_dict_mismatch():
    given = {
        "Activity Name - 3.9.1": "autoclaved aerated concrete block production",
        "Geography - 3.9.1": "IN",
        "Reference Product - 3.9.1": "autoclaved aerated concrete block;\nhard coal ash",
        "Reference Product Unit - 3.9.1": "kg",
        "Activity Name - 3.10": "autoclaved aerated concrete block production",
        "Geography - 3.10": "IN",
        "Reference Product - 3.10": "autoclaved aerated concrete block;\nhard coal ash",
        "Reference Product Unit - 3.10": "kg;\nkg",
    }
    with pytest.raises(Mismatch):
        source_target_pair_as_bw_dict(given, "3.9.1", "3.10")


def test_source_target_pair_as_bw_dict_uncombinable():
    given = {
        "Activity Name - 3.9.1": "autoclaved aerated concrete block production",
        "Geography - 3.9.1": "IN",
        "Reference Product - 3.9.1": "autoclaved aerated concrete block;\nhard coal ash;\nsomething",
        "Reference Product Unit - 3.9.1": "kg;\nkg;\nMJ",
        "Activity Name - 3.10": "autoclaved aerated concrete block production",
        "Geography - 3.10": "IN",
        "Reference Product - 3.10": "autoclaved aerated concrete block;\nhard coal ash",
        "Reference Product Unit - 3.10": "kg;\nkg",
    }
    with pytest.raises(Uncombinable):
        source_target_pair_as_bw_dict(given, "3.9.1", "3.10")


def test_source_target_pair_as_bw_dict_valueerror():
    given = {
        "Activity Name - 3.9.1": "autoclaved aerated concrete block production",
        "Geography - 3.9.1": "IN",
        "Reference Product - 3.9.1": "autoclaved aerated concrete block;\nhard coal ash",
        "Reference Product Unit - 3.9.1": "kg;\nkg",
        "Activity Name - 3.10": "autoclaved aerated concrete block production",
        "Geography - 3.10": "IN",
        "Reference Product - 3.10": "autoclaved aerated concrete block;\nhard coal ash",
        "Reference Product Unit - 3.10": "kg;\nkg",
    }
    assert source_target_pair_as_bw_dict(given, "3.9.1", "3.10")
    with pytest.raises(ValueError):
        source_target_pair_as_bw_dict(given, "3.8", "3.10")
    with pytest.raises(ValueError):
        source_target_pair_as_bw_dict(given, "3.9.1", "3.11")


def test_source_target_pair_as_bw_dict_new_dataset():
    given = {
        "Activity Name - 3.9.1": float("NaN"),
        "Geography - 3.9.1": float("NaN"),
        "Reference Product - 3.9.1": float("NaN"),
        "Reference Product Unit - 3.9.1": float("NaN"),
        "Activity Name - 3.10": "baling",
        "Geography - 3.10": "GLO",
        "Reference Product - 3.10": "baling",
        "Reference Product Unit - 3.10": "unit",
    }
    assert source_target_pair_as_bw_dict(given, "3.9.1", "3.10") == []


def test_source_target_pair_as_bw_dict_multiple_some_missing():
    given = {
        "Activity Name - 3.9.1": "p-nitrotoluene production",
        "Geography - 3.9.1": "GLO",
        "Reference Product - 3.9.1": "nan;\nnan;\np-nitrotoluene",
        "Reference Product Unit - 3.9.1": "nan;\nnan;\nkg",
        "Activity Name - 3.10": "nitrotoluenes production, toluene nitration",
        "Geography - 3.10": "GLO",
        "Reference Product - 3.10": "m-nitrotoluene;\no-nitrotoluene;\np-nitrotoluene",
        "Reference Product Unit - 3.10": "kg;\nkg;\nkg",
    }
    expected = [
        {
            "source": {
                "name": "p-nitrotoluene production",
                "location": "GLO",
                "reference product": "p-nitrotoluene",
                "unit": "kg",
            },
            "target": {
                "name": "nitrotoluenes production, toluene nitration",
                "location": "GLO",
                "reference product": "p-nitrotoluene",
                "unit": "kg",
            },
        },
    ]
    assert source_target_pair_as_bw_dict(given, "3.9.1", "3.10") == expected



"""ecoinvent_migrate."""

__all__ = (
    "__version__",
    "generate_technosphere_mapping",
    "generate_biosphere_mapping",
)

__version__ = "0.2.0"

from .main import generate_biosphere_mapping, generate_technosphere_mapping



import json
import os
from datetime import datetime, timezone
from pathlib import Path
from typing import List, Optional

from loguru import logger
from platformdirs import user_data_dir

from . import __version__


def write_datapackage(
    data: dict,
    source_db_name: str,
    target_db_name: str,
    licenses: Optional[List[dict]] = None,
    output_directory: Optional[Path] = None,
    version: str = "1.0.0",
    description: Optional[str] = None,
) -> Path:
    now = datetime.now(timezone.utc).isoformat()
    if licenses is None:
        licenses = [
            {
                "name": "CC BY 4.0",
                "path": "https://creativecommons.org/licenses/by/4.0/",
                "title": "Creative Commons Attribution 4.0 International",
            }
        ]
    if description is None:
        description = f"Data migration file from {source_db_name} to {target_db_name} generated with `ecoinvent_migrate` version {__version__}"

    formatted = {
        "name": f"{source_db_name}-{target_db_name}",
        "licenses": licenses,
        "version": version,
        "description": description,
        "homepage": "https://github.com/brightway-lca/ecoinvent_migrate",
        "created": now,
        "contributors": [
            {"title": "ecoinvent association", "path": "https://ecoinvent.org/", "role": "author"},
            {"title": "Chris Mutel", "path": "https://chris.mutel.org/", "role": "wrangler"},
        ],
    } | data

    if output_directory is None:
        output_directory = Path(user_data_dir("ecoinvent_migrate", "pylca"))
        output_directory.mkdir(parents=True, exist_ok=True)
    elif not isinstance(output_directory, Path):
        raise ValueError(
            f"`output_directory` must be a `Path` instance; got {type(output_directory)}"
        )
    elif not output_directory.is_dir() or not os.access(output_directory, os.W_OK):
        raise ValueError(f"`output_directory` {output_directory} must be a writable directory")

    fp = output_directory / f"{source_db_name}-{target_db_name}.json"
    logger.info("Writing output file {fp}", fp=str(fp))
    with open(fp, "w") as f:
        json.dump(formatted, f, indent=2, ensure_ascii=False)

    return fp



import itertools
import math
from collections import defaultdict
from numbers import Number
from typing import List, Union

from loguru import logger

from .errors import Mismatch, Uncombinable


def isnan(o: Union[str, Number]) -> bool:
    return isinstance(o, Number) and math.isnan(o)


def split_by_semicolon(row: dict, version: str) -> list[dict]:
    """Possible split a data row into"""
    if isinstance(row[f"Activity Name - {version}"], Number) and math.isnan(
        row[f"Activity Name - {version}"]
    ):
        return []

    len_product = len(row[f"Reference Product - {version}"].split(";\n"))
    len_unit = len(row[f"Reference Product Unit - {version}"].split(";\n"))
    if len_product != len_unit:
        raise Mismatch(f"Can't match {len_product} products to {len_unit} units")

    return [
        {
            "name": row[f"Activity Name - {version}"],
            "location": row[f"Geography - {version}"],
            "reference product": x,
            "unit": y,
        }
        for x, y in zip(
            row[f"Reference Product - {version}"].split(";\n"),
            row[f"Reference Product Unit - {version}"].split(";\n"),
        )
    ]


def source_target_pair_as_bw_dict(
    row: dict, source_version: str, target_version: str
) -> list[dict]:
    """Transform a row from the change report dataframe into one or more dicts.

    Can be more than one because the change report is for master data which can have
    multiple reference products.

    Only include the attributes we use for linking:

    * name
    * reference product
    * unit
    * location

    Example usage:

    ```python
    >>> given = {
        'Activity Name - 3.9.1': 'baling',
        'Geography - 3.9.1': 'GLO',
        'Reference Product - 3.9.1': 'baling',
        'Reference Product Unit - 3.9.1': 'unit',
        'Activity Name - 3.10': 'baling',
        'Geography - 3.10': 'GLO',
        'Reference Product - 3.10': 'baling',
        'Reference Product Unit - 3.10': 'unit',
    }
    >>> source_target_pair_as_bw_dict(given)
    [{
        'source': {
            'name': 'baling',
            'location': 'GLO',
            'reference product': 'baling',
            'unit': 'unit'
        },
        'target': {
            'name': 'baling',
            'location': 'GLO',
            'reference product': 'baling',
            'unit': 'unit'
        }
    }]
    ```

    But multiple pairs are returned when more than one reference product is present:

    ```python
    >>> given = {
        'Activity Name - 3.9.1': 'autoclaved aerated concrete block production',
        'Geography - 3.9.1': 'IN',
        'Reference Product - 3.9.1': 'autoclaved aerated concrete block;\nhard coal ash',
        'Reference Product Unit - 3.9.1': 'kg;\nkg',
        'Activity Name - 3.10': 'autoclaved aerated concrete block production',
        'Geography - 3.10': 'IN',
        'Reference Product - 3.10': 'autoclaved aerated concrete block;\nhard coal ash',
        'Reference Product Unit - 3.10': 'kg;\nkg',
    }
    >>> source_target_pair_as_bw_dict(given)
    [{
        'source': {
            'name': 'autoclaved aerated concrete block production',
            'location': 'IN',
            'reference product': 'autoclaved aerated concrete block',
            'unit': 'kg'
        },
        'target': {
            'name': 'autoclaved aerated concrete block production',
            'location': 'IN',
            'reference product': 'autoclaved aerated concrete block',
            'unit': 'kg'
        }
    }, {
        'source': {
            'name': 'autoclaved aerated concrete block production',
            'location': 'IN',
            'reference product': 'hard coal ash',
            'unit': 'kg'
        },
        'target': {
            'name': 'autoclaved aerated concrete block production',
            'location': 'IN',
            'reference product': 'hard coal ash',
            'unit': 'kg'
        }
    }]
    ```

    """
    versions = [x.split(" - ")[-1].strip() for x in row if x.startswith("Activity Name")]
    if f"Activity Name - {source_version}" not in row:
        raise ValueError(
            f"""Can't find source version {source_version} in data row.
    Versions found: {versions}"""
        )
    if f"Activity Name - {target_version}" not in row:
        raise ValueError(
            f"""Can't find target version {target_version} in data row.
    Versions found: {versions}"""
        )

    sources = split_by_semicolon(row, source_version)
    if not sources:
        # New unit process dataset, no source objects
        return []

    targets = split_by_semicolon(row, target_version)
    if len(sources) > 1 and len(targets) > 1 and len(sources) != len(targets):
        raise Uncombinable(
            f"Can't do M:N combination of {len(sources)} source and {len(targets)} target datasets."
        )
    elif len(sources) == 1 and len(targets) == 1:
        return [{"source": sources[0], "target": targets[0]}]
    elif len(sources) == 1:
        sources = itertools.repeat(sources[0])
    elif len(targets) == 1:
        targets = itertools.repeat(targets[0])

    return [
        {"source": s, "target": t}
        for s, t in zip(sources, targets)
        if all(v.lower() != "nan" for v in itertools.chain(s.values(), t.values()))
    ]


def resolve_glo_row_rer_roe(
    data: List[dict],
    source_db_name: str,
    target_db_name: str,
    source_lookup: dict,
    target_lookup: dict,
) -> List[dict]:
    """Iterate through `data`, and change `location` attribute to `RoW` or `RoE` when needed.

    Looks in actual database to get correct `location` attributes."""

    for obj in data:
        source_missing = None
        for kind, lookup, db_name in [
            ("source", source_lookup, source_db_name),
            ("target", target_lookup, target_db_name),
        ]:
            key = tuple([obj[kind][attr] for attr in ("name", "location", "reference product")])
            if key in lookup:
                continue
            elif (
                key not in lookup
                and obj[kind]["location"] == "GLO"
                and (key[0], "RoW", key[2]) in lookup
            ):
                obj[kind]["location"] = "RoW"
                logger.debug(
                    "{kind} process {name} location corrected to 'RoW'",
                    kind=kind,
                    name=obj[kind]["name"],
                )
            elif (
                key not in lookup
                and obj[kind]["location"] == "RER"
                and (key[0], "RoE", key[2]) in lookup
            ):
                obj[kind]["location"] = "RoE"
                logger.debug(
                    "{kind} process {name} location corrected to 'RoE'",
                    kind=kind,
                    name=obj[kind]["name"],
                )
            else:
                if kind == "target" and source_missing:
                    # Missing in both source and target for this system model
                    source_missing = None
                    continue
                elif kind == "source":
                    source_missing = obj[kind]
                else:
                    # Only missing in target database - but this is a big problem, we don't have a
                    # suitable target for existing edges to relink to.
                    logger.warning(
                        "{kind} process given in change report but missing in {db_name} lookup: {ds}",
                        kind=kind.title(),
                        db_name=db_name,
                        ds=obj[kind],
                    )
        if source_missing:
            # Only a debug message because this won't break anything - there is no process in the
            # source database to miss a link from.
            logger.debug(
                "Source process given in change report but missing in {db_name} lookup: {ds}",
                db_name=source_db_name,
                ds=source_missing,
            )

    return data


FIELDS = ("name", "location", "reference product", "unit")


def astuple(obj: dict, fields: List[str] = FIELDS) -> tuple:
    return tuple([obj[field] for field in fields])


def disaggregated(data: List[dict], lookup: dict) -> dict:
    """Take a list of mapping dictionaries with the same `source`, and create one `disaggregate`
    object.

    Applies `allocation` factors based on the production volumes in `lookup`

    """
    for obj in data:
        try:
            obj["pv"] = lookup[
                astuple(obj["target"], ("name", "location", "reference product"))
            ].rp_exchange()["production volume"]
        except KeyError:
            logger.warning(
                """Change report annex dataset missing from database.
    This is likely a publication error which you can't fix.
    Removing the following from migrations:
    {ds}""",
                ds=obj["target"],
            )
            obj["pv"] = 0

    total = sum(obj["pv"] for obj in data)
    if not total:
        logger.warning(
            "Total production from {n} targets is zero for source {s}; using equal allocation factors",
            n=len(data),
            s=data[0]["source"],
        )
        return {
            "source": data[0]["source"],
            "targets": [obj["target"] | {"allocation": 1 / len(data)} for ob in data],
        }
    elif total < 0:
        logger.warning(
            "Total production from {n} targets is less than zero for source {s}; what is happening!?",
            n=len(data),
            s=data[0]["source"],
        )

    return {
        "source": data[0]["source"],
        "targets": [obj["target"] | {"allocation": obj["pv"] / total} for obj in data if obj["pv"]],
    }


def split_replace_disaggregate(data: List[dict], target_lookup: dict) -> dict:
    """Split the transformations in `data` into `replace` and `disaggregate` sections.

    Disaggregation is needed when one dataset is replaced by multiple datasets. We lookup the
    respective production volumes to get the disaggregation factors."""
    groupie = defaultdict(list)
    for obj in data:
        groupie[astuple(obj["source"])].append(obj)

    return {
        "replace": [
            value[0]
            for value in groupie.values()
            if len(value) == 1 and value[0]["source"] != value[0]["target"]
        ],
        "disaggregate": [
            disaggregated(value, target_lookup) for value in groupie.values() if len(value) > 1
        ],
    }


def get_column_labels(example: dict, version: str) -> dict:
    """Guess column labels from Excel change report annex."""
    uuid_tries = [f"UUID - {version}", f"ID - {version}"]
    for uuid_try in uuid_tries:
        if uuid_try in example:
            uuid = uuid_try
            break
    else:
        raise ValueError(f"Can't find uuid field for database version {version} in {example}")
    name_tries = [f"Name - {version}", f"{version} name", f"{version} - name"]
    for name_try in name_tries:
        if name_try in example:
            name = name_try
            break
    else:
        raise ValueError(f"Can't find name field for database version {version} in {example}")
    return {
        "uuid": uuid,
        "name": name,
    }


def source_target_biosphere_pair(
    data: List[dict], source_version: str, target_version: str, keep_deletions: bool
) -> List[dict]:
    """Turn pandas DataFrame rows into source/target pairs."""
    source_labels = get_column_labels(example=data[0], version=source_version)
    target_labels = get_column_labels(example=data[0], version=target_version)

    formatted = {
        "replace": [
            {
                "source": {k: row[v] for k, v in source_labels.items()},
                "target": {k: row[v] for k, v in target_labels.items()},
                "allocation": float(row.get("Conversion Factor (old-new)", 1.0)),
                "comment": row.get("Comment"),
            }
            for row in data
            if not isnan(row[target_labels["uuid"]])
        ]
    }
    if keep_deletions:
        formatted["delete"] = [
            {
                "source": {k: row[v] for k, v in source_labels.items()},
                "comment": row.get("Comment"),
            }
            for row in data
            if isnan(row[target_labels["uuid"]])
        ]

    for lst in formatted.values():
        for obj in lst:
            if "comment" in obj and (not obj["comment"] or isnan(obj["comment"])):
                del obj["comment"]
            if "allocation" in obj and (obj["allocation"] == 1.0 or isnan(obj["allocation"])):
                del obj["allocation"]

    return formatted



import warnings
from pathlib import Path
from typing import List, Optional

import pandas as pd
from ecoinvent_interface import EcoinventRelease, Settings
from loguru import logger

from .data_io import get_brightway_databases, get_change_report_filepath, setup_project
from .datapackage import write_datapackage
from .errors import VersionJump
from .utils import configure_logs
from .wrangling import (
    resolve_glo_row_rer_roe,
    source_target_biosphere_pair,
    source_target_pair_as_bw_dict,
    split_replace_disaggregate,
)


def get_change_report_context(
    source_version: str,
    target_version: str,
    project_name: str,
    system_model: str = "cutoff",
    ecoinvent_username: Optional[str] = None,
    ecoinvent_password: Optional[str] = None,
) -> Path:
    """Get the source/target change report filepath, and setup Brightway project with needed data.

    Source and target must be one release apart from each other, i.e. 3.4 to 3.5 or 3.7 to 3.7.1.

    Returns the `Path` of the generated file."""
    if ecoinvent_username is not None or ecoinvent_password is not None:
        warnings.warn(
            """
Username and/or password supplied via function call.
Please consider setting these permanently. See:
https://github.com/brightway-lca/ecoinvent_interface?tab=readme-ov-file#authentication-via-settings-object
            """
        )
        settings = Settings(username=ecoinvent_username, password=ecoinvent_password)
    else:
        settings = Settings()
    release = EcoinventRelease(settings)
    versions = release.list_versions()

    if source_version not in versions:
        raise ValueError(
            f"Given source version {source_version} not in available versions: {versions}"
        )
    if target_version not in versions:
        raise ValueError(
            f"Given target version {source_version} not in available versions: {versions}"
        )

    logger.info("Versions available for this license: {versions}", versions=versions)

    excel_filepath = get_change_report_filepath(version=target_version, release=release)
    if source_version not in excel_filepath.name.replace(target_version, ""):
        raise VersionJump(
            f"""
Source ({source_version}) and target ({target_version}) don't have a change report.
Usually this is the case when one jumps across multiple releases, but not always.
For example, the change report for 3.7.1 is from 3.6, not 3.7.
The change report we have available is:
{excel_filepath.name}
        """
        )
    logger.info("Using change report annex file {filename}", filename=excel_filepath.name)

    setup_project(
        source_version=source_version,
        target_version=target_version,
        project_name=project_name,
        system_model=system_model,
        ecoinvent_username=ecoinvent_username,
        ecoinvent_password=ecoinvent_password,
    )
    return excel_filepath


def generate_technosphere_mapping(
    source_version: str,
    target_version: str,
    project_name: str = "ecoinvent-migration",
    system_model: str = "cutoff",
    ecoinvent_username: Optional[str] = None,
    ecoinvent_password: Optional[str] = None,
    write_logs: bool = True,
    licenses: Optional[List[dict]] = None,
    output_directory: Optional[Path] = None,
    output_version: str = "1.0.0",
    description: Optional[str] = None,
) -> Path:
    """Generate a Randonneur mapping file for technosphere edge attributes from source to target."""
    configure_logs(write_logs=write_logs)

    excel_filepath = get_change_report_context(
        source_version=source_version,
        target_version=target_version,
        project_name=project_name,
        ecoinvent_username=ecoinvent_username,
        ecoinvent_password=ecoinvent_password,
    )

    sheet_names = pd.ExcelFile(excel_filepath).sheet_names
    candidates = [name for name in sheet_names if name.lower() == "qualitative changes"]
    if not candidates:
        raise ValueError(
            "Can't find suitable sheet name in change report file. Looking for 'qualitative changes', found:\n\t{}".format(
                "\n\t".join(sheet_names)
            )
        )
    elif len(candidates) > 1:
        raise ValueError(
            "Found multiple sheet names like 'qualitative changes' for change report file:\n\t{}".format(
                "\n\t".join(sheet_names)
            )
        )

    data = [
        pair
        for row in pd.read_excel(io=excel_filepath, sheet_name=candidates[0]).to_dict(
            orient="records"
        )
        for pair in source_target_pair_as_bw_dict(row, source_version, target_version)
    ]

    source_db_name, target_db_name, source_lookup, target_lookup = get_brightway_databases(
        source_version=source_version, target_version=target_version, system_model=system_model
    )
    data = resolve_glo_row_rer_roe(
        data=data,
        source_db_name=source_db_name,
        target_db_name=target_db_name,
        source_lookup=source_lookup,
        target_lookup=target_lookup,
    )
    data = split_replace_disaggregate(data=data, target_lookup=target_lookup)

    if not data["replace"] and not data["disaggregate"]:
        logger.info(
            "It seems like there are no technosphere changes for this release. Doing nothing."
        )
        return
    if not data["replace"]:
        del data["replace"]
    if not data["disaggregate"]:
        del data["disaggregate"]

    return write_datapackage(
        data=data,
        source_db_name=source_db_name,
        target_db_name=target_db_name,
        licenses=licenses,
        output_directory=output_directory,
        version=output_version,
        description=description,
    )


def generate_biosphere_mapping(
    source_version: str,
    target_version: str,
    keep_deletions: bool = False,
    project_name: str = "ecoinvent-migration",
    ecoinvent_username: Optional[str] = None,
    ecoinvent_password: Optional[str] = None,
    write_logs: bool = True,
    licenses: Optional[List[dict]] = None,
    output_directory: Optional[Path] = None,
    output_version: str = "1.0.0",
    description: Optional[str] = None,
) -> Path:
    """Generate a Randonneur mapping file for biosphere edge attributes from source to target."""
    configure_logs(write_logs=write_logs)

    logger.info(
        """The `EE Deletions` format is not consistent across versions.
Please check the outputs carefully before applying them."""
    )

    excel_filepath = get_change_report_context(
        source_version=source_version,
        target_version=target_version,
        project_name=project_name,
        ecoinvent_username=ecoinvent_username,
        ecoinvent_password=ecoinvent_password,
    )

    sheet_names = pd.ExcelFile(excel_filepath).sheet_names
    candidates = [name for name in sheet_names if name.lower() == "ee deletions"]
    if not candidates:
        logger.info(
            "It seems like there are no biosphere changes; no sheet name like `EE Deletions` found. Sheet names found:\n\t{sn}",
            sn="\n\t".join(sheet_names),
        )
        return
    elif len(candidates) > 1:
        raise ValueError(
            "Found multiple sheet names like 'EE Deletions' for change report file:\n\t{}".format(
                "\n\t".join(sheet_names)
            )
        )

    data = pd.read_excel(io=excel_filepath, sheet_name=candidates[0]).to_dict(orient="records")
    data = source_target_biosphere_pair(
        data=data,
        source_version=source_version,
        target_version=target_version,
        keep_deletions=keep_deletions,
    )

    return write_datapackage(
        data=data,
        source_db_name=f"ecoinvent-{source_version}-biosphere",
        target_db_name=f"ecoinvent-{target_version}-biosphere",
        licenses=licenses,
        output_directory=output_directory,
        version=output_version,
        description=description,
    )



from pathlib import Path
from typing import Optional

import bw2data as bd
import bw2io as bi
from ecoinvent_interface import EcoinventRelease
from loguru import logger

from .errors import MissingDatabase


def get_change_report_filepath(version: str, release: EcoinventRelease) -> Path:
    """Get the filepath to the Excel change report file"""
    files = release.list_extra_files(version)
    candidates = [key for key in files if "change report" in key.lower() and "annex" in key.lower()]
    if not candidates:
        raise ValueError(
            "Can't find suitable change report filename from release files:\n\t{}".format(
                "\n\t".join([f.name for f in files])
            )
        )
    elif len(candidates) > 1:
        raise ValueError(
            "Found multiple candidates for change report filename:\n\t{}".format(
                "\n\t".join([c.name for c in candidates])
            )
        )

    files = list(release.get_extra(version, candidates[0]).iterdir())
    candidates = [
        fp
        for fp in files
        if (
            fp.name.lower().startswith("change report annex")
            or fp.name.lower().startswith("change_report_annex")
        )
        and fp.suffix.lower() == ".xlsx"
    ]
    if not candidates:
        raise ValueError(
            "Can't find suitable change report file from archive:\n\t{}".format(
                "\n\t".join([f.name for f in files])
            )
        )
    elif len(candidates) > 1:
        raise ValueError(
            "Found multiple candidates for change report file:\n\t{}".format(
                "\n\t".join([c.name for c in candidates])
            )
        )
    return candidates[0]


def setup_project(
    source_version: str,
    target_version: str,
    project_name: str,
    system_model: str = "cutoff",
    ecoinvent_username: Optional[str] = None,
    ecoinvent_password: Optional[str] = None,
) -> None:
    bd.projects.set_current(project_name)

    if f"ecoinvent-{source_version}-{system_model}" not in bd.databases:
        logger.info(
            "Downloading ecoinvent source version {version} {system_model} to resolve GLO/RoW and RER/RoE",
            version=source_version,
            system_model=system_model,
        )
        bi.import_ecoinvent_release(
            source_version,
            system_model,
            lcia=False,
            username=ecoinvent_username,
            password=ecoinvent_password,
        )
    if f"ecoinvent-{target_version}-{system_model}" not in bd.databases:
        logger.info(
            "Downloading ecoinvent target version {version} {system_model} to resolve GLO/RoW and RER/RoE",
            version=target_version,
            system_model=system_model,
        )
        bi.import_ecoinvent_release(
            target_version,
            system_model,
            lcia=False,
            username=ecoinvent_username,
            password=ecoinvent_password,
        )


def get_brightway_databases(source_version: str, target_version: str, system_model: str):
    source_db_name = f"ecoinvent-{source_version}-{system_model}"
    target_db_name = f"ecoinvent-{target_version}-{system_model}"
    if source_db_name not in bd.databases:
        raise MissingDatabase(f"Missing source database: {source_db_name}")
    if target_db_name not in bd.databases:
        raise MissingDatabase(f"Missing target database: {target_db_name}")

    logger.info("Loading source database {db} to cache data attributes", db=source_db_name)
    source_lookup = {
        tuple([o[attr] for attr in ("name", "location", "reference product")]): o
        for o in bd.Database(source_db_name)
    }
    logger.info("Loading target database {db} to cache data attributes", db=target_db_name)
    target_lookup = {
        tuple([o[attr] for attr in ("name", "location", "reference product")]): o
        for o in bd.Database(target_db_name)
    }
    return source_db_name, target_db_name, source_lookup, target_lookup



import datetime
import sys
from pathlib import Path

from loguru import logger
from platformdirs import user_log_dir


def configure_logs(write_logs: bool = True) -> None:
    logger.remove()
    logger.add(sys.stderr, level="INFO")
    if write_logs:
        logs_dir = Path(
            user_log_dir("ecoinvent_migrate", "pylca")
        ) / datetime.datetime.now().isoformat()[:19].replace(":", "-")
        logger.info("Writing logs to {path}", path=logs_dir)
        logs_dir.mkdir(parents=True, exist_ok=True)
        logger.add(logs_dir / "debug.log", level="DEBUG")
        logger.add(logs_dir / "info.log", level="INFO")
